# Copyright (c) 2008, David Baird <dhbaird@gmail.com>
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. Neither the name of the <ORGANIZATION> nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

'''
====================================================
Splarnektity: Functional Pattern Matching for Python
====================================================

Splarnektity provides a limited, but valuable, form of functional pattern
matching to Python, a la OCaml, Erlang, Haskell, et al..  Another way
of thinking of this is: pattern matching over Python lists.

.. sectnum::

.. contents:: Contents

History
=======

2008-08-19 16:00:24 UTC  David Baird <dhbaird@gmail.com>

  Version 0.3 release.  This release fixes some important bugs with
  ``.others`` (there was an off-by-one errors that prevented matching
  of zero-length tuples; and there was a missing check that could all
  matching tuples that shouldn't match).  Regression tests have been
  written for these bugs, and a few more units tests have been added.

2008-08-09 23:59:10 UTC  David Baird <dhbaird@gmail.com>

  Version 0.2 release.  This release is a documentation and example
  update.  The vars() method was added to provide a more consistent
  interface (just in case it is ever needed) than var().

2008-07-24 00:51:13 UTC  David Baird <dhbaird@gmail.com>

  First release, including documentation (in the module docstring)
  and some test cases.  I originally wanted to use the name FMatch, but
  that was already taken;  So were PFMatch and FunMatch.  I had to draw
  the line at PyMatch and PyFMatch.  Those names just don't sound cool.
  In the end, I gave up and just decided to name the module Splarnektity
  (inspired by Ze Frank).

License
=======

The license for Splarnektity is a standard BSD license.  See the comments
at the top of the source code for the license information.

Motivation
==========

This library was originally born from a motivation to process sexps
in Python.  S-expressions (aka sexps) are a concise way to represent
structured data that both humans and computers can interpret.  A sexp
might look like::

    '(netlist
        (namespace connector (main-board J2))
        (namespace fpga (main-board U17))
        (net (connector 1) (fpga AE12)) 
        (net (connector 2) (fpga AA22)) 
        (net (connector 3) (connector 4) (fpga B32)) 
    )

See, easy for humans to grasp, right?  But what about computers?

Python (Version 2.5) versus Functional
--------------------------------------

Let's examine a family of [currently] related languages: Python, Ruby,
C, and Java, but with emphasis on Python.  We're going to assume that
the sexp has been converted to a Python-style list or tuple already and
thus looks like this::

    s = ('quote',
      ('netlist',
        ('namespace', 'connector', ('main-board', 'J2')),
        ('namespace', 'fpga', ('main-board', 'U17')),
        ('net', ('connector', '1'), ('fpga', 'AE12')),
        ('net', ('connector', '2'), ('fpga', 'AA22')),
        ('net', ('connector', '3'), ('connector', '4'), ('fpga', 'B32')) 
      )
    )

Phew!  That's a lot of commas and quotes!  Here's how one would normally
process this statement in Python::

    assert(len(s) == 2 and s[0] == 'quote')
    netlist = s[1]
    assert(len(netlist) > 0 and netlist[0] == 'netlist')
    for item in netlist:
        assert(len(item) > 0)
        if item[0] == 'namespace':
            assert(len(item) == 3)
            name, components = item[1], item[2]
            print 'namespace', name, components
        if item[0] == 'net':
            assert(len(item) > 1)
            connections = item[1:]
            print 'net',
            for namespace, port in connections:
                print '%s:%s' % (namespace, port),
            print

There are several annoying aspects about this code.  The first annoying
aspect is that the the programmer has to keep checking the size of lists
before accessing items in the list.  The second annoying thing is that
before unpacking a list, the programmer has to be certain what type of
list is about to be unpacked.  Functional pattern matching solves these
annoyances, freeing the programmer's mind to focus on more important
things!

Here is an example of what functional pattern matching looks like when
it is officially supported by a language::

    match s:
      when ('quote', netlist):
        match netlist:
          when ('netlist', *items):
            for item in items:
              match item:
                when ('namespace', name, components):
                  print 'namespace', name, components
                when ('net', *connections):
                  print 'net', 
                  for connection in connections:
                    match connection:
                      when (namespace, port):
                        print '%s:%s' % (namespace, port),
                  print

I'm going to highlight a portion of each code to demonstrate where
functional pattern matching really starts to pay off: when multiple
cases must be handled.  Look more closely at this bit of Python which
has more than one case::

    ...
        if item[0] == 'namespace':
            assert(len(item) == 3)
            name, components = item[1], item[2]
            ...
        if item[0] == 'net':
            assert(len(item) > 1)
            connections = item[1:]
            ...

and also look at the pseudo code for these cases::

    ...
                match item:
                  when ('namespace', name, components):
                    ...
                  when ('net', *connections):
                    ...

Notice that the comparison, unpacking, and length checking all happen in
one statement (the ``when`` statement) in the pseudo code.  To be fair,
the Python code could be compacted a little bit more, but it is still too
verbose and the programmer still must worry about indexing and unpacking::

    ...
        if len(item) == 3 and item[0] == 'namespace':
            name, components = item[1], item[2]
        if len(item) > 1 and item[0] == 'net':
            connections = item[1:]
            ...

So far, the pseudo code looks kinda good.

What else could we ask for?  How about nested patterns and also
conditions.  Okay, here they are::

    match s:
      when ('quote', ('netlist', *items)), len(items) > 0: # <- look here!!
        for item in items:
          match item:
            when ('namespace', name, components):
              print 'namespace', name, components
            when ('net', *connections):
              print 'net', 
              for connection in connections:
                match connection:
                  when (namespace, port):
                    print '%s:%s' % (namespace, port),
              print


How about less indentation?::

    match s:
    when ('quote', ('netlist', *items)), len(items) > 0:
        for item in items:
          match item:
          when ('namespace', name, components):
              print 'namespace', name, components
          when ('net', *connections):
              print 'net', 
              for connection in connections:
                match connection:
                when (namespace, port):
                    print '%s:%s' % (namespace, port),
              print


A Real FMatch Example
---------------------

The following pseudo code::

    match s:
      when ('quote', ('netlist', *items)), len(items) > 0:
        for item in items:
          match item:
            when ('namespace', name, components):
              print 'namespace', name, components
            when ('net', *connections):
              print 'net', 
              for connection in connections:
                match connection:
                when (namespace, port):
                    print '%s:%s' % (namespace, port),
              print

translates to the following real Python code with FMatch::

    import splarnektity as fmatch

    s = ('quote',
      ('netlist',
        ('namespace', 'connector', ('main-board', 'J2')),
        ('namespace', 'fpga', ('main-board', 'U17')),
        ('net', ('connector', '1'), ('fpga', 'AE12')),
        ('net', ('connector', '2'), ('fpga', 'AA22')),
        ('net', ('connector', '3'), ('connector', '4'), ('fpga', 'B32')) 
      )
    )

    f = fmatch.FMatch(s)
    items = f.var()

    if f.when(('quote', ('netlist', items.others))) and len(items.v) > 0:
        for item in items.v:
            f = fmatch.FMatch(item)
            name, components, connections = f.var(3)
            if False: pass
            elif f.when(('namespace', name, components)):
                print 'namespace', name.v, components.v
            elif f.when(('net', connections.others)):
                print 'net',
                for connection in connections.v:
                    g = fmatch.FMatch(connection)
                    namespace, port = g.var(2)
                    if g.when((namespace, port)):
                        print '%s:%s' % (namespace.v, port.v),
                print

The result of executing this program is::

    namespace connector ('main-board', 'J2')
    namespace fpga ('main-board', 'U17')
    net connector:1 fpga:AE12
    net connector:2 fpga:AA22
    net connector:3 connector:4 fpga:B32

Caveats
=======

Splarnektity (i.e. Reality) versus The Ideal
--------------------------------------------

Aesthetics
~~~~~~~~~~

Here's a bit of the ideal pseudo code::

    ...
          match item:
            when ('namespace', name, components):
              print 'namespace', name, components
            when ('net', *connections):
              ...
            else:
              ...

and here's the reality code::

    ...
            f = fmatch.FMatch(item)
            name, components, connections = f.var(3)
            if   f.when(('namespace', name, components)):
                print 'namespace', name.v, components.v
            elif f.when(('net', connections.others)):
                ...
            else:
              ...

The obvious difference is that Python has no match/when keywords (anyone
interested in taking up an effort to fix this?? ^_^).  Also, all of the
variables must be declared in advance in the real code (i.e. ``name, ... =
f.var(3)``) which is kind of a pain.  Also, accessing vars in the real
code requires accessing the ``.v`` member of the var.  To get the value of
a var, the ``.v`` accessor (i.e. ``name.v``) is used.  These methods have
pros and cons.  The pro is that all sorts of cool methods (like ``.int``
to check if the type of the variable is an int or long) can be added to
the pattern/match objects.  The con is that ``name.v`` looks ugly.

To summarize, the aesthetic shortcomings of Splarnektity are:

- Variables must be instantiated in advance
- The value of a variable must be accessed via the ``.v`` accessor method

Performance
~~~~~~~~~~~

The performance of Splarnektity is surely not as great as achieved with
a system that actual compiles patterns into a pattern matching automaton.

Python Tuple Syntax
~~~~~~~~~~~~~~~~~~~

This is mentioned in the Python docs, but in case someone doesn't already
know, here is how to create Tuple of *one* item within Python::

    (1,)

The comma is required.  Otherwise, the parentheses becomes an
order-of-precedence kind of operator (rather than a tuple syntax).

FMatch Examples and Applications
================================

A Simplistic FMatch Example
---------------------------

Example usage::

    from splarnektity import FMatch, X

    value = ('do not care', (17, 'happy'), 12)
    f = fmatch.FMatch(value)
    a = f.var()

    if f.when((X, (17, 'happy'), a)) and a.int and a.v <= 2:
        print a.v # >>> (NOT REACHED)

    if f.when((X, (17, 'sad'), a)) and a.int and a.v > 2:
        print a.v # >>> (NOT REACHED)

    if f.when((X, (17, 'happy'), a)) and a.int and a.v > 2:
        print a.v # >>> 12

    if f.when((X, (17, 'happy'), a.I)) and a.v > 2:
        print a.v # >>> 12

Term Rewriting
--------------

Caveat: I'll be impressed if an efficient term rewriting system can be
made with Splarnektity.  Actually, using Splarnektity for large scale
term rewriting is probably a bad idea.  If a better term rewriting
system is needed, these languages/tools are worth looking into:

- http://q-lang.sourceforge.net/ Q - Equational Programming Language
- http://pure-lang.sourceforge.net/ The Pure Programming Language
- ML (OCaml)
- Maybe even this: http://www.program-transformation.org/Stratego/
- ...and surely others...

Q (and its successor, Pure) is pretty fun though and I recommend checking
it out.

With that said, here is an example of a very simple term rewriter based
on Splarnektity:

    `examples/simple_term_rewriting.html <examples/simple_term_rewriting.html>`_
'''

class FVar(object):

    def __init__(self, context, ref=None):
        self._ctxt = context
        if ref is None: self._ref = self
        else:           self._ref = ref

    @property
    def v(self): # v is for value
        '''Returns the current value assigned to the variable (by calling
FMatch.where(...). If the variable has not been assigned yet,
then NoValue is returned.'''
        return self._ctxt.get_var(self)

    @property
    def int(self):
        '''Returns true if the value assigned to this FVar is an int'''
        return type(self.v) in [int, long]

    @property
    def float(self):  return type(self.v) == float

    @property
    def num(self):    return type(self.v) in [float, int, long]

    @property
    def str(self):    return type(self.v) == str

    @property
    def list(self):   return type(self.v) in [tuple, list]

    @property
    def I(self):
        '''When used in pattern matching, this forces an integer constraint
on the matching process.  Matching will fail if the value is not
an integer.'''
        return FVarI(self._ctxt, self._ref)

    @property
    def F(self):      return FVarF(self._ctxt, self._ref)

    @property
    def N(self):      return FVarN(self._ctxt, self._ref)

    @property
    def S(self):      return FVarS(self._ctxt, self._ref)

    @property
    def L(self):      return FVarL(self._ctxt, self._ref)

    @property
    def others(self): return FVarOthers(self._ctxt, self._ref)

    def set(self, value):
        if self._ctxt: self._ctxt.set_var(self, value)

    def test(self, value):  return True

class FVarI(FVar):
    def test(self, value): return type(value) in [int, long]
class FVarF(FVar):
    def test(self, value): return type(value) == float
class FVarN(FVar):
    def test(self, value): return type(value) in [float, int, long]
class FVarS(FVar):
    def test(self, value): return type(value) == str
class FVarL(FVar):
    def test(self, value): return type(value) in [list, tuple]
class FVarOthers(FVar):
    def test(self, value): return True

class NoValue(object): pass

#class DoNotCare(FVar): pass
DoNotCare = FVar(None)
X = DoNotCare # convenient alias

class FMatch(object):

    def __init__(self, value):
        self.value = value
        self.clear_vars()

    def var(self, count=0):
        '''Factory method for creating FVars'''
        if count == 0: return FVar(self)
        else:          return [FVar(self) for x in xrange(count)]

    def vars(self, count=0):
        '''Similar to the var() method, but always returns a list of
FVar objects for the sake of consistency (even when only 1 var
is requested).'''
        return [FVar(self) for x in xrange(count)]

    def get_var(self, which_var):
        if which_var in self._vars: return self._vars[which_var._ref]
        else:                       return NoValue

    def set_var(self, which_var, value):
        self._vars[which_var._ref] = value

    def clear_vars(self):
        self._vars = { }

    def when(self, pattern):
        self.clear_vars()
        return self._when(pattern, self.value)

    def _when(self, pattern, value):

        tl = [tuple, list]
        if type(pattern) in tl:
            if type(self.value) in tl:
                # This was the old buggy line of code:
                #if len(pattern) > 1 and type(pattern[-1]) == FVarOthers:
                if len(pattern) > 0 \
                   and type(pattern[-1]) == FVarOthers \
                   and len(self.value) >= (len(pattern) - 1):
                    for p, v in zip(pattern[:-1], value):
                        if not self._when(p, v): return False
                    assert(pattern[-1]._ctxt == self)
                    pattern[-1].set(value[len(pattern)-1:])
                elif len(pattern) == len(value):
                    for p, v in zip(pattern, value):
                        if not self._when(p, v): return False
                else:
                    return False
            else:
                return False
            return True

        fil = [float, int, long]
        if type(pattern) in fil and type(value) in fil:
            if pattern != value: return False
            else:                return True

        if type(pattern) == str and type(value) == str:
            if pattern != value: return False
            else:                return True

        if type(pattern) in [FVar, FVarI, FVarF, FVarN, FVarS, FVarL]:
            if pattern._ctxt:
                assert(pattern._ctxt == self)
                pattern.set(value)
            else: # DoNotCare
                pass
            if pattern.test(value): return True
            else:                   return False

        if pattern == DoNotCare:
            return True

        return False

def test():

    if True:

        value = ('do not care', (17, 'happy'), 12)
        f = FMatch(value)
        a = f.var()

        if f.when((X, (17, 'happy'), a)) and a.int and a.v <= 2:
            print a.v # >>> (skipped)
            assert(False)

        if f.when((X, (17, 'sad'), a)) and a.int and a.v > 2:
            print a.v # >>> (skipped)
            assert(False)

        if f.when((X, (17, 'happy'), a)) and a.int and a.v > 2:
            print a.v # >>> 12
            assert(True)

        if f.when((X, (17, 'happy'), a.I)) and a.v > 2:
            print a.v # >>> 12
            assert(True)

        if f.when((X, (17, 'happy'), a.S)):
            print a.v # >>> (skipped)
            assert(False)

    if True:

        value = ('netlist', ('net', 1, 2), ('net', 5, 4, 3))
        f = FMatch(value)
        nets = f.var()

        if f.when(('netlist', nets.others)):
            for net in nets.v:
                f = FMatch(net)
                ports = f.var()
                if f.when(('net', ports.others)) and len(ports.v) > 0:
                    print ports.v

        # Non-pattern matching (i.e. vanilla Python)
        if value[0] == 'netlist':
            nets = value[1:]
            for net in nets:
                if net[0] == 'net':
                    ports = net[1:]
                    print ports
    if True:
        s = ('quote',
          ('netlist',
            ('namespace', 'connector', ('main-board', 'J2')),
            ('namespace', 'fpga', ('main-board', 'U17')),
            ('net', ('connector', '1'), ('fpga', 'AE12')),
            ('net', ('connector', '2'), ('fpga', 'AA22')),
            ('net', ('connector', '3'), ('connector', '4'), ('fpga', 'B32')) 
          )
        )

        f = FMatch(s)
        items = f.var()

        if f.when(('quote', ('netlist', items.others))) and len(items.v) > 0:

            for item in items.v:
                f = FMatch(item)
                name, components, connections = f.var(3)

                if False: pass

                elif f.when(('namespace', name, components)):
                    print 'namespace', name.v, components.v

                elif f.when(('net', connections.others)):

                    print 'net',
                    for connection in connections.v:
                        g = FMatch(connection)
                        namespace, port = g.var(2)

                        if g.when((namespace, port)):
                            print '%s:%s' % (namespace.v, port.v),
                    print

    if True:
        f = FMatch((1, 2, 3, 4))
        Z = f.var()

        assert(f.when((Z.others,)));              assert(Z.v == (1, 2, 3, 4))
        assert(f.when(Z));                        assert(Z.v == (1, 2, 3, 4))
        assert(f.when((1, Z.others,)));           assert(Z.v == (2, 3, 4))
        assert(f.when((1, 2, 3, Z.others,)));     assert(Z.v == (4,))
        assert(f.when((1, 2, 3, 4, Z.others,)));  assert(Z.v == ())
        assert(not f.when((2, Z.others,)))

    if True:
        f = FMatch(())
        Z = f.var()

        assert(f.when((Z.others,)));  assert(Z.v == ())
        assert(f.when(()))
        assert(f.when(Z));            assert(Z.v == ())
        assert(not f.when((X,)))
        assert(not f.when((2,)))
        assert(not f.when((2, 3, Z.others,)))
        assert(not f.when((2, Z.others,)))


if __name__ == '__main__':
    test()
